# -*- coding: utf-8 -*-
# Author: Yiping Liu
# Description: This script calculates the average of a list of numbers.
# Version: 1.0
# Last Modified: May 7, 2023


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from scipy import stats

# Load the CSV file
data = pd.read_csv(r'E:\work\data\low-T\error_analysis.csv')

# Extract the data columns
RK = data['RK']
Tm = data['Tm']
mNa = data['mNa']

# Plot the histograms
fig, axs = plt.subplots(1, 3, figsize=(10, 4))

axs[0].hist(RK, color='blue')
axs[0].set_xlabel('RK')
axs[0].set_ylabel('Count')

axs[1].hist(Tm, color='green')
axs[1].set_xlabel('Tm')
axs[1].set_ylabel('Count')

axs[2].hist(mNa, color='red')
axs[2].set_xlabel('mNa')
axs[2].set_ylabel('Count')

# Calculate the 68% confidence interval for mNa
conf_int = stats.norm.interval(0.68, loc=np.mean(mNa), scale=np.std(mNa) / np.sqrt(len(mNa)))

print(conf_int)
# Calculate the uncertainty as half of the range of the confidence interval
uncertainty = (conf_int[1] - conf_int[0]) / 2

# Plot the confidence interval
axs[2].axvspan(conf_int[0], conf_int[1], alpha=0.5, color='grey')

# Show the plot
plt.show()

plt.savefig(r'E:\work\plots\error histogram.pdf', format='pdf')
mNa_mean = np.mean(mNa)

# Print the mean and uncertainty as x ± δx
print('Mean mNa: {:.3f} ± {:.3f}'.format(mNa_mean, uncertainty))
# Save the figure to a PDF file



# -*- coding: utf-8 -*-
# Author: Yiping Liu
# Description: This script calculates the average of a list of numbers.
# Version: 1.0
# Last Modified: May 7, 2023


import sys
import os
import numpy as np

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.dirname(SCRIPT_DIR))

from Pitzer.model2 import FluidPitzer

# shape = (0, 2)
# result_array = np.empty(shape, dtype=np.float64)
# ts =[
# -20
# ]
# for t in ts:
#     fluid = FluidPitzer(
#         x0=(3, 1),
#         species=['Na+', 'Ca+2'],
#         t=t + 273.16,
#         solids=['NaCl-2H2O', 'H2O(S)'],
#     )
#     print(fluid.get_molalities([1,2]))
#     result = fluid.optimize()
#     if result.success:
#         print(result.x)
#         result_array = np.append(result_array, [result.x], axis=0)
# print(result_array)


fluid = FluidPitzer(
        x0=(3, 1),
        species=['Na+', 'Ca+2'],
        t=-22 + 273.16,
        solids=['NaCl-2H2O', 'H2O(S)'],
    )

print(fluid.optimize())


# -*- coding: utf-8 -*-
# Author: Yiping Liu
# Description:
# Version: 1.0
# Last Modified: May 7, 2023

import sys
import os
import pprint

pp = pprint.PrettyPrinter(width=41, compact=True)

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.dirname(SCRIPT_DIR))

from sympy import *
import uncertainty.methods as pm
import itertools

from uncertainty.methods import (
    get_charge_number,
)

"""
Assuming the molality of ion "Na+" is x;
and the molality of ion "Cl-" is y.
"""

x, y, z = symbols('x,y,z')


class FluidPitzer:
    def __init__(self, dictionary=None, t=25, database='lassin',):
        if dictionary is None:
            dictionary = {}
        self.dictionary = {}
        sum = 0
        for key, value in dictionary.items():
            if "-" not in key and "+" not in key:
                self.dictionary[key] = value
            elif '-' not in key:
                self.dictionary[key] = value * x
                sum += value * x
            self.dictionary[key] = value
        self.dictionary['Cl-'] = sum
        self.t = t + 273.16
        self.database = database

    def get_molalities(self):
        return self.dictionary

    def charge_balance(self):
        balance = 0
        for species in self.dictionary.keys():
            if '-' in species or '+' in species:
                balance += get_charge_number(species) * self.dictionary[species]
        return balance

    def get_total_ratios_of_cations(self):
        cations = self.dictionary.copy()
        r_total = 0
        if 'Cl-' or 'SO4-2' in cations:
            cations.pop('Cl-')
            cations.pop('SO4-2')
        for key, value in cations.items():
            charge_number = get_charge_number(key)
            r_total += charge_number * value
        return r_total / x

    def get_ionic_strength(self):
        return pm.get_ionic_strength(self.get_molalities())

    def get_z(self):
        molalities = self.get_molalities()
        ions = molalities.keys()
        z_value = 0
        for ion in ions:
            charge_number = pm.get_charge_number(ion)
            z_value += molalities[ion] * abs(charge_number)
        return z_value

    def get_component_groups(self):
        components = self.get_molalities().keys()
        groups = pm.group_components(components)
        return groups

    def get_a_phi(self):
        return pm.a_phi_moller(self.t)

    def get_b(self):
        pair_parameters = {}
        cation_anion_pairs = self.get_component_groups()['cation_anion_pairs']
        for cap in cation_anion_pairs:
            pair_parameters[cap] = pm.beta_calculate(
                ion_pair=cap,
                ionic_strength=self.get_ionic_strength(),
                t=self.t
            )
        return pair_parameters

    def get_b_phi(self):
        pair_parameters = {}
        cation_anion_pairs = self.get_component_groups()['cation_anion_pairs']
        for cap in cation_anion_pairs:
            pair_parameters[cap] = pm.beta_phi_calculate(
                ion_pair=cap,
                ionic_strength=self.get_ionic_strength(),
                t=self.t
            )

        return pair_parameters

    def get_b_prime(self):

        """
        Get Betas ready for calculating the "F" function.
        :return: betas for each type of salt
        """

        pair_parameters = {}
        cation_anion_pairs = self.get_component_groups()['cation_anion_pairs']

        for cap in cation_anion_pairs:
            pair_parameters[cap] = pm.beta_prime_calculate(
                ion_pair=cap,
                ionic_strength=self.get_ionic_strength(),
                t=self.t
            )
        return pair_parameters

    def get_c(self):
        pair_parameters = {}
        cation_anion_pairs = self.get_component_groups()['cation_anion_pairs']
        for cap in cation_anion_pairs:
            if 'Li+' in cap or 'LiCl0' in cap:
                method = 'lassin'
            else:
                method = 'moller'
            pair_parameters[cap] = pm.c_calculate(
                ion_pair=cap,
                t=self.t,
            )
        return pair_parameters

    def get_cc_phi(self):
        cation_pairs = self.get_component_groups()['cation_pairs']
        phi_dict = {}
        for cation_pair in cation_pairs:
            phi_dict[cation_pair] = pm.cc_phi_calculate(
                ion_pair=cation_pair,
                ionic_strength=self.get_ionic_strength(),
                a_phi=self.get_a_phi(),
                t=self.t
            )
        return phi_dict

    def get_cc_phi_prime_phi(self):
        i = self.get_ionic_strength()
        dict = {}
        phis = self.get_cc_phi()

        for pair in phis.keys():
            dict[pair] = phis[pair]["phi"] + i * phis[pair]["phi_prime"]
        return dict

    def get_aa_phi(self):
        if self.get_component_groups()['anion_pairs']:
            anion_pairs = self.get_component_groups()['anion_pairs']
            phi_dict = {}
            for anion_pair in anion_pairs:
                phi_dict[anion_pair] = pm.aa_phi_calculate(
                    ion_pair=anion_pair,
                    ionic_strength=self.get_ionic_strength(),
                    a_phi=self.get_a_phi(),
                    t=self.t
                )
            return phi_dict
        return 'Anions less than 1'

    def get_aa_phi_prime_phi(self):
        if self.get_component_groups()['anion_pairs']:
            i = self.get_ionic_strength()
            phi_prime_phi_dict = {}
            phis = self.get_aa_phi()
            for pair in phis.keys():
                phi_prime_phi_dict[pair] = phis[pair]["phi"] + i * phis[pair]["phi_prime"]
            return phi_prime_phi_dict

    def get_cca_psi(self):
        cation_pairs = self.get_component_groups()['cation_pairs']
        anions = self.get_component_groups()['anions']
        cca_pairs = {}
        for cation_pair in cation_pairs:
            cation1 = cation_pair[0]
            cation2 = cation_pair[1]
            for anion in anions:
                rd = pm.ternary_parameters_ready((cation1, cation2, anion), self.t)
                cca_pairs[cation1, cation2, anion] = pm.get_parameter(pair=(cation1, cation2, anion), name='psi',
                                                                      data=rd, t=self.t)
        return cca_pairs

    def get_aac_psi(self):
        if self.get_component_groups()['anion_pairs']:
            anion_pairs = self.get_component_groups()['anion_pairs']
            cations = self.get_component_groups()['cations']
            aac_pairs = {}
            for anion_pair in anion_pairs:
                anion_list = list(anion_pair)
                anion1 = anion_list[0]
                anion2 = anion_list[1]
                for cation in cations:
                    rd = pm.ternary_parameters_ready((anion1, anion2, cation), self.t)
                    aac_pairs[anion1, anion2, cation] = pm.get_parameter(pair=(anion1, anion2, cation), name='psi',
                                                                         data=rd, t=self.t)
            return aac_pairs

    def get_lambdas(self):
        neutral_ion_pairs = self.get_component_groups()['neutral_ion_pairs']
        lambdas = {}
        for pair in neutral_ion_pairs:
            rd = pm.binary_parameters_ready(pair, self.t)
            lambda_value = pm.get_parameter(pair=pair, name='lambda', data=rd, t=self.t)
            lambdas[pair] = lambda_value
        return lambdas

    def get_zetas(self):
        nca_pairs = self.get_component_groups()['neutral_cation_anion_pairs']
        zetas = {}
        for pair in nca_pairs:
            rd = pm.ternary_parameters_ready(pair, self.t)
            zeta_value = pm.get_parameter(pair=pair, name='zeta', data=rd, t=self.t)
            zetas[pair] = zeta_value
        return zetas

    def get_osmotic_coefficient(self):
        groups = self.get_component_groups()
        molalities = self.get_molalities()
        m_sum = sum(molalities.values())
        i = self.get_ionic_strength()
        a_phi = self.get_a_phi()
        z = self.get_z()
        b = self.get_b_phi()
        c = self.get_c()
        cations = groups['cations']
        cation_pairs = groups['cation_pairs']
        anions = groups['anions']
        anion_pairs = groups['anion_pairs']
        cc_phis = self.get_cc_phi_prime_phi()
        cca_psis = self.get_cca_psi()

        neutrals = groups['neutrals']
        lambdas = self.get_lambdas()
        zetas = self.get_zetas()
        nca_pairs = groups['neutral_cation_anion_pairs']

        item0 = (2 / m_sum)

        # print('item0:', item0)

        item1 = -(a_phi * i ** 1.5) / (1 + 1.2 * i ** 0.5)

        item2 = 0
        for cap in b.keys():
            m_1 = molalities[cap[0]]
            m_2 = molalities[cap[1]]

            b_phi = 0
            for key in b.keys():
                if set(cap) == set(key):
                    b_phi = b[key]

            c_value = 0
            for key in c.keys():
                if set(cap) == set(key):
                    c_value = c[key]

            item2 += m_1 * m_2 * (b_phi + z * c_value)

        item3 = 0

        for cp in cation_pairs:
            c1 = cp[0]
            c2 = cp[1]

            m_c1 = molalities[c1]
            m_c2 = molalities[c2]

            # find value of cc_phi
            cc_phi = 0
            for key in cc_phis.keys():
                if {cp[0], cp[1]} == set(key):
                    cc_phi = cc_phis[key]

            item3_subitem1 = 0
            for anion in anions:
                m_a = molalities[anion]

                # find value of cca_psi
                cca_psi = 0
                for key in cca_psis.keys():
                    if set(key) == {c1, c2, anion}:
                        cca_psi = cca_psis[key]

                item3_subitem1 += m_a * cca_psi

            item3 += m_c1 * m_c2 * (cc_phi + item3_subitem1)

        item4 = 0

        if len(anions) > 1:
            aa_phis = self.get_aa_phi_prime_phi()
            aac_psis = self.get_aac_psi()

            for ap in anion_pairs:
                m_a1 = molalities[ap[0]]
                m_a2 = molalities[ap[1]]

                # find value of aa_phi
                aa_phi = 0
                for key in aa_phis.keys():
                    if {ap[0], ap[1]} == set(key):
                        aa_phi = aa_phis[key]

                item4_subitem1 = 0
                for cation in cations:
                    m_c = molalities[cation]

                    aac_psi = 0

                    # find value of aac_psi
                    for key in cca_psis.keys():
                        if {ap[0], ap[1], cation} == set(key):
                            aac_psi = aac_psis[key]

                    item4_subitem1 += m_c * aac_psi
                item4 += m_a1 * m_a2 * (aa_phi + item4_subitem1)

        item5 = 0
        for neutral in neutrals:
            m_n = molalities[neutral]
            for cation in cations:
                m_c = molalities[cation]

                lambda_nc = 0
                for key in lambdas.keys():
                    if {neutral, cation} == set(key):
                        lambda_nc = lambdas[key]
                item5 += m_n * m_c * lambda_nc

        item6 = 0
        for neutral in neutrals:
            m_n = molalities[neutral]
            for anion in anions:
                m_a = molalities[anion]

                lambda_na = 0
                for key in lambdas.keys():
                    if {neutral, anion} == set(key):
                        lambda_na = lambdas[key]
                item6 += m_n * m_a * lambda_na

        item7 = 0
        for nca_pair in nca_pairs:
            m1 = molalities[nca_pair[0]]
            m2 = molalities[nca_pair[1]]
            m3 = molalities[nca_pair[2]]

            zeta_nca = 0
            for key in zetas.keys():
                if set(nca_pair) == set(key):
                    zeta_nca = zetas[key]

            item7 += m1 * m2 * m3 * zeta_nca

        osmotic_coefficient = item0 * (item1 + item2 + item3 + item4 + item5 + item6 + item7) + 1
        return osmotic_coefficient

    def get_water_activity(self):
        m_sum = sum(self.get_molalities().values())
        # print('m_sum:', m_sum)
        phi = self.get_osmotic_coefficient()
        ln_activity = -phi * m_sum / 55.50844
        return ln_activity

    def get_f_uppercase(self):
        a_phi = self.get_a_phi()
        i = self.get_ionic_strength()
        component_groups = self.get_component_groups()
        cation_pairs = component_groups['cation_pairs']
        anion_pairs = component_groups['anion_pairs']
        molalities = self.get_molalities()
        ca_beta_primes = self.get_b_prime()
        cation_anion_pairs = component_groups['cation_anion_pairs']
        cc_phis = self.get_cc_phi()
        aa_phis = self.get_aa_phi()
        f_gamma = pm.get_f_gamma(a_phi, i)

        item0 = f_gamma

        item1 = 0
        for cap in cation_anion_pairs:
            m_p1 = molalities[cap[0]]
            m_p2 = molalities[cap[1]]
            ca_beta_prime = 0

            # find the value of beta_prime of cap
            for key in ca_beta_primes.keys():
                if {cap[0], cap[1]} == set(key):
                    ca_beta_prime = ca_beta_primes[key]

            item1 += m_p1 * m_p2 * ca_beta_prime

        item2 = 0
        for cp in cation_pairs:
            m_p1 = molalities[cp[0]]
            m_p2 = molalities[cp[1]]

            cc_phi_prime = 0

            # find the value of beta_prime of cap
            for key in cc_phis.keys():
                if {cp[0], cp[1]} == set(key):
                    cc_phi_prime = cc_phis[key]['phi_prime']

            item2 += m_p1 * m_p2 * cc_phi_prime

        item3 = 0
        if anion_pairs:
            for ap in anion_pairs:
                m_a1 = molalities[ap[0]]
                m_a2 = molalities[ap[1]]

                aa_phi_prime = 0
                for key in aa_phis.keys():
                    if {ap[0], ap[1]} == set(key):
                        aa_phi_prime = aa_phis[key]['phi_prime']

                item3 += m_a1 * m_a2 * aa_phi_prime

        f_uppercase = item0 + item1 + item2 + item3
        return f_uppercase

    def get_cation_activity_coefficients(self, target_cation):
        groups = self.get_component_groups()
        cations = groups['cations']
        anions = groups['anions']
        anion_pairs = groups['anion_pairs']
        cation_anion_pairs = groups['cation_anion_pairs']
        neutrals = groups['neutrals']
        molalities = self.get_molalities()
        target_cation = target_cation
        betas = self.get_b()
        cs = self.get_c()
        z = self.get_z()
        lambdas = self.get_lambdas()
        zetas = self.get_zetas()
        cc_phis = self.get_cc_phi()
        cation_psis = self.get_cca_psi()
        anion_psis = self.get_aac_psi()

        f_uppercase = self.get_f_uppercase()
        charge_number = get_charge_number(target_cation)

        item0 = charge_number ** 2 * f_uppercase

        item1 = 0
        for anion in anions:
            m_a = molalities[anion]
            beta_ca = 0
            for key in betas.keys():
                if {target_cation, anion} == set(key):
                    beta_ca = betas[key]
            c_ca = 0
            for key in cs.keys():
                if {target_cation, anion} == set(key):
                    c_ca = cs[key]
            item1 += m_a * (2 * beta_ca + z * c_ca)

        item2 = 0
        for cation in cations:
            if cation != target_cation:
                m_c = molalities[cation]
                phi_mc = 0
                for key in cc_phis.keys():
                    if set(key) == {cation, target_cation}:
                        phi_mc = cc_phis[key]['phi']

                item2_subitem1 = 0
                for anion in anions:
                    m_a = molalities[anion]
                    for tkey in cation_psis.keys():
                        if {target_cation, cation, anion} == set(tkey):
                            item2_subitem1 += m_a * cation_psis[tkey]
                item2 += m_c * (2 * phi_mc + item2_subitem1)

        item3 = 0
        if anion_pairs:
            for ap in anion_pairs:
                m_a1 = molalities[ap[0]]
                m_a2 = molalities[ap[1]]
                psi_maa = 0

                # find match for psi
                for key in anion_psis.keys():
                    if {ap[0], ap[1], target_cation} == set(key):
                        psi_maa = anion_psis[key]

                item3 += m_a1 * m_a2 * psi_maa

        item4 = 0
        for cap in cation_anion_pairs:
            ion1 = cap[0]
            ion2 = cap[1]
            c_ca = 0

            # find match for c_ca
            for key in cs.keys():
                if set(key) == {ion1, ion2}:
                    c_ca = cs[key]

            item4 += molalities[ion1] * molalities[ion2] * c_ca

        item5 = 0
        for neutral in neutrals:
            m_n = molalities[neutral]

            lambda_nm = 0
            for key in lambdas.keys():
                if {neutral, target_cation} == set(key):
                    lambda_nm = lambdas[key]
            item5 += m_n * lambda_nm

        item6 = 0
        for neutral in neutrals:
            m_n = molalities[neutral]
            for anion in anions:
                m_a = molalities[anion]

                zeta_nam = 0
                for key in zetas.keys():
                    if {target_cation, anion, neutral} == set(key):
                        zeta_nam = zetas[key]
                item6 += m_n * m_a * zeta_nam

        ln_coefficient = item0 + item1 + item2 + item3 + abs(charge_number) * item4 + 2 * item5 + item6
        return ln_coefficient

    def get_anion_activity_coefficients(self, target_anion):
        groups = self.get_component_groups()
        cations = groups['cations']
        anions = groups['anions']
        cation_pairs = groups['cation_pairs']
        cation_anion_pairs = groups['cation_anion_pairs']
        neutrals = groups['neutrals']
        lambdas = self.get_lambdas()
        zetas = self.get_zetas()
        molalities = self.get_molalities()
        target_anion = target_anion
        betas = self.get_b()
        cs = self.get_c()
        z = self.get_z()

        aa_phis = self.get_aa_phi()

        aac_psis = self.get_aac_psi()
        cca_psis = self.get_cca_psi()

        f_uppercase = self.get_f_uppercase()
        charge_number = get_charge_number(target_anion)

        item0 = charge_number ** 2 * f_uppercase

        item1 = 0
        for cation in cations:
            m_a = molalities[cation]
            beta_ca = 0
            for key in betas.keys():
                if {target_anion, cation} == set(key):
                    beta_ca = betas[key]
            c_ca = 0
            for key in cs.keys():
                if {target_anion, cation} == set(key):
                    c_ca = cs[key]
            item1 += m_a * (2 * beta_ca + z * c_ca)

        item2 = 0
        for anion in anions:
            if anion != target_anion:
                m_a = molalities[anion]
                phi_xa = 0
                for key in aa_phis.keys():
                    if set(key) == {anion, target_anion}:
                        phi_xa = aa_phis[key]['phi']

                item2_subitem1 = 0
                for cation in cations:
                    m_c = molalities[cation]
                    for tkey in aac_psis.keys():
                        if {target_anion, cation, anion} == set(tkey):
                            item2_subitem1 += m_c * aac_psis[tkey]
                item2 += m_a * (2 * phi_xa + item2_subitem1)

        item3 = 0
        for cp in cation_pairs:
            m_c1 = molalities[cp[0]]
            m_c2 = molalities[cp[1]]
            psi_ccx = 0

            # find match for psi
            for key in cca_psis.keys():
                if {cp[0], cp[1], target_anion} == set(key):
                    psi_ccx = cca_psis[key]
            item3 += m_c1 * m_c2 * psi_ccx

        item4 = 0
        for cap in cation_anion_pairs:
            ion1 = cap[0]
            ion2 = cap[1]
            m1 = molalities[ion1]
            m2 = molalities[ion2]

            c_ca = 0
            for key in cs.keys():
                if set(key) == {ion1, ion2}:
                    c_ca = cs[key]

            item4 += m1 * m2 * c_ca

        item5 = 0
        for neutral in neutrals:
            m_n = molalities[neutral]
            lambda_nx = 0

            for key in lambdas.keys():
                if {neutral, target_anion} == set(key):
                    lambda_nx = lambdas[key]

            item5 += m_n * lambda_nx

        item6 = 0
        for neutral in neutrals:
            m_n = molalities[neutral]
            for cation in cations:
                m_c = molalities[cation]

                zeta_nam = 0

                for key in zetas.keys():
                    if {target_anion, cation, neutral} == set(key):
                        zeta_nam = zetas[key]
                item6 += m_n * m_c * zeta_nam

        ln_coefficient = item0 + item1 + item2 + item3 + abs(charge_number) * item4 + 2 * item5 + item6

        return ln_coefficient

    def get_neutral_activity_coefficients(self, target_neutral):
        groups = self.get_component_groups()
        cations = groups['cations']
        anions = groups['anions']
        cation_anion_pairs = groups['cation_anion_pairs']

        molalities = self.get_molalities()

        lambdas = self.get_lambdas()
        zetas = self.get_zetas()

        item0 = 0
        for cation in cations:
            m_c = molalities[cation]
            lambda_nc = 0
            for key in lambdas.keys():
                if {cation, target_neutral} == set(key):
                    lambda_nc = lambdas[key]
            item0 += m_c * lambda_nc

        item1 = 0
        for anion in anions:
            m_a = molalities[anion]
            lambda_na = 0
            for key in lambdas:
                if {anion, target_neutral} == set(key):
                    lambda_na = lambdas[key]
            item1 += m_a * lambda_na

        item2 = 0
        for pair in cation_anion_pairs:
            m1 = molalities[pair[0]]
            m2 = molalities[pair[1]]

            zeta_nca = 0
            for key in zetas.keys():
                if {target_neutral, pair[0], pair[1]} == set(key):
                    zeta_nca = zetas[key]

            item2 += m1 * m2 * zeta_nca

        ln_coefficient = 2 * item0 + 2 * item1 + item2

        return ln_coefficient





import sys
import os
import numpy as np
import matplotlib.pyplot as plt

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.dirname(SCRIPT_DIR))

from MCWEP.models import FluidPitzer

# Set the parameters
a = 2,
b = 25
# Set the uncertainties in the parameters
da = 0.28
db = 0.2
# Define the number of Monte Carlo simulations to run
N = 1000
# Define an array to store the results
x_values = np.zeros(N)

# Define arrays to store the values of a_rand and b_rand
a_rand = np.zeros(N)
b_rand = np.zeros(N)

# Perform the Monte Carlo simulation
for i in range(N):
    # Generate random values for the parameters based on their uncertainties
    a_rand[i] = np.random.normal(a, da)
    b_rand[i] = np.random.normal(b, db)
    # Use the find_x function to find the value of x given the random parameters
    # print('a_rand:',a_rand)
    fluid = FluidPitzer(
        x0=0.1,
        species={
            'Na+': 1,
            'K+': a_rand[i],
        },
        t=b_rand[i],
        solids=['KCl']
    )

    x_values[i] = fluid.find_x()
    print('finished:', i)

x_mean = np.mean(x_values)
x_std = np.std(x_values)

np.savetxt('E:\\work\\data\\low-T\\error_analysis.csv', np.column_stack((a_rand, b_rand, x_values)), delimiter=',',
           header='RK,Tm,mNa', comments='')

# # Plot the histograms of a_rand, b_rand, and x_values
# fig, axs = plt.subplots(1, 3, figsize=(12, 4))
# axs[0].hist(a_rand, bins=20)
# axs[0].set_xlabel('RK')
# axs[0].set_ylabel('Frequency')
# axs[1].hist(b_rand, bins=20)
# axs[1].set_xlabel('Tm')
# axs[1].set_ylabel('Frequency')
# axs[2].hist(x_values, bins=20)
# axs[2].set_xlabel('mNa')
# axs[2].set_ylabel('Frequency')
# plt.tight_layout()
# plt.show()

# Print the results
print("Mean value of x: ", x_mean)
print("Standard deviation of x: ", x_std)



# -*- coding: utf-8 -*-
# Author: Yiping Liu
# Description: This script calculates the average of a list of numbers.
# Version: 1.0
# Last Modified: May 7, 2023

import sys
import os

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.dirname(SCRIPT_DIR))

from Pitzer.models import FluidPitzer
import pandas as pd


# columns = [
#     'T',
#     'Na+',
#     'Ca+2',
#     # 'K+',
#     # 'Li+',
#     # 'Mg+2',
#     # 'Sr+2',
#     # 'Cs+',
#     # 'Rb+',
#     'Mn+2',
#     'Zn+2',
#     'Pb+2'
# ]
# df = pd.DataFrame(columns=columns)
#
# def replace_blanks_with_commas(input_string):
#     # replace blank spaces with commas in each row and convert to tuple of floats (or strings)
#     rows = input_string.strip().split('\n')
#     replaced_rows = []
#     for row in rows:
#         row_as_list = []
#         for num in row.split():
#             try:
#                 row_as_list.append(float(num))
#             except ValueError:
#                 row_as_list.append(num)
#         row_as_tuple = tuple(row_as_list)
#         replaced_rows.append(row_as_tuple)
#
#     # output as list of tuples
#     output = replaced_rows
#
#     return output
#
#
# data = """
# -21.90364364	1	0.261712115	0.002102761	0.002044225	0.000300444
# """
# data = replace_blanks_with_commas(data)
#
#
#
# for dt in data:
#     print(dt)
#     species = {
#         'Na+':  dt[1],
#         'Ca+2': dt[2],
#         # 'K+':   dt[3],
#         # 'Li+':  dt[4],
#         # 'Mg+2': dt[5],
#         # 'Sr+2': dt[6],
#         # 'Cs+':  dt[7],
#         # 'Rb+':  dt[8],
#         'Mn+2': dt[3],
#         'Zn+2': dt[4],
#         'Pb+2': dt[5]
#     }
#     fluid = FluidPitzer(
#         x0=(3, 5),
#         species=species,
#         t=dt[0],
#         solids=['H2O(S)']
#     )
#     print(fluid.solids)
#     result = fluid.optimize()
#     if result.success:
#         for key, value in species.items():
#             species[key] = value * result.x[0]
#
#     values_list = list(species.values())
#
#     # create a new DataFrame with the current row data and concatenate it with the existing DataFrame
#     row_data = [dt[0]] + values_list
#     row_df = pd.DataFrame([row_data], columns=columns)
#     df = pd.concat([df, row_df], ignore_index=True)
#
# print(df)
#
# # write the DataFrame to a CSV file
# df.to_csv(r'E:\work\data\low-T\general_result.csv', index=False)

# aqueous species determined in LA-ICP-MS analysis
species = {
    'Na+': 1,
    'K+':  2,
}
fluid = FluidPitzer(
    x0=(3, 3),
    species=species,
    # melting temperature of the last solid
    t=25,
    # the last melting solid
    solids=['KCl']
)

result = fluid.optimize()

print(result)


# -*- coding: utf-8 -*-
# Author: Yiping Liu
# Description: This script calculates the average of a list of numbers.
# Version: 1.0
# Last Modified: May 7, 2023

import numpy as np
import pandas as pd
import itertools
from sympy import *

from functools import lru_cache


from database.marion_binary import marion_binary
from database.marion_ternary import marion_ternary


from database.spencer_revised_binary import spencer_binary
from database.spencer_revised_ternary import spencer_ternary
from database.spencer_revised_chemical_potential import spencer_chemical_potential_db
from database.marion_chemical_potential import marion_chemical_potential_db
from database.lassin_chemical_potential import lassin_chemical_potential_db
from database.solid_data import solids
from public.low_level import find_pair
from database.lassin_binary import lassin_binary
from database.lassin_ternary import lassin_ternary
from database.lassin_chemical_potentials import lassin_chemical_potential_db
from database.solid_data import solids
from public.low_level import find_pair
from public.j_x import compute_j_jp

import functools


# to make dictionary harshable
def hash_dict(func):
    """Transform mutable dictionnary
    Into immutable
    Useful to be compatible with cache
    """

    class HDict(dict):
        def __hash__(self):
            return hash(frozenset(self.items()))

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        args = tuple([HDict(arg) if isinstance(arg, dict) else arg for arg in args])
        kwargs = {k: HDict(v) if isinstance(v, dict) else v for k, v in kwargs.items()}
        return func(*args, **kwargs)

    return wrapped


def get_two_ion_db(name):
    name = name.lower()
    database = None
    if name == 'spencer':
        database = spencer_binary
    elif name == 'marion':
        database = marion_binary
    elif name == 'lassin':
        database = lassin_binary
    return database


def get_three_ion_db(name):
    name = name.lower()
    database = None
    if name == 'spencer':
        database = spencer_ternary
    elif name == 'marion':
        database = marion_ternary
    return database


@hash_dict
@lru_cache(maxsize=None)
def find_parameter_value(pair, dbname, parameter_name, col_name):
    """
    Find values for parameters from database based on ion names.
    :param pair: a tuple of ion pair.
    :param dbname: database name.
    :param parameter_name: the name of the parameter, e.g. 'b0', 'b1' ...
    :param col_name:
    :return:
    """
    database = get_two_ion_db(dbname)

    first_level_df = database.loc[parameter_name]
    ion_pairs = first_level_df.index

    parameter = 0
    for key in ion_pairs:
        if {pair[0], pair[1]} == set(key):
            parameter = first_level_df.loc[key, col_name]

    if isinstance(parameter, pd.Series):
        parameter = parameter.values[0]

    return float(parameter)


def binary_parameters_ready(pair, t, dbname):
    """
    Make the A0 - A5 value parameters ready in a dictionary for further selection and calculating
    :param pair: a tuple of ion pair.
    :param t: temperature.
    :return: values of parameters.
    """
    if 'Fe+2' in pair:
        dbname = 'marion'

    database = get_two_ion_db(dbname)

    # get the number of columns
    table_width = len(database.columns)

    parameter_names = ['b0', 'b1', 'b2', 'c_phi', 'theta', 'lambda']
    dic = {}
    for pn in parameter_names:
        for col in database.columns[:table_width]:
            value = find_parameter_value(pair, dbname, pn, col)
            dic[pn + '_' + col] = value
    return dic


def ternary_parameters_ready(pair, t, dbname):
    """
    Make parameters ready in a dictionary for further selection and calculating
    :param pair: A tuple a three ions.
    :return: Interaction parameters of this group of ions.
    """
    if 'Fe+2' in pair:
        dbname = 'marion'

    database = get_three_ion_db(dbname)

    pns = ['psi', 'zeta']
    parameters = {}
    for pn in pns:
        first_level_df = database.loc[pn]
        ion_pairs = first_level_df.index
        col_list = database.columns

        match = find_pair(pair, ion_pairs)
        if match['has']:
            target = match['target']
            for col in col_list:
                parameter = first_level_df.loc[target, col]
                parameters[pn + '_' + col] = parameter
        else:
            for col in col_list:
                parameter = 0
                parameters[pn + '_' + col] = parameter
    return parameters


def chemical_potential_lassin(a1, a2, a3, a4, a5, t):
    log_k = a1 + a2 * t + a3 / t + a4 * np.log10(t) + a5 / (t ** 2)
    return log_k * 2.303


def parameter_cal_lassin(a1, a2, a3, a4, a5, a6, a7, a8, t):
    p = a1 + a2 * t + a3 * t ** 2 + a4 * t ** 3 + a5 / t + a6 * ln(t) + a7 / (t - 263) + a8 / (680 - t)
    return p


def parameter_cal_spencer(a1, a2, a3, a4, a5, a6, t):
    parameter = a1 + a2 * t + a3 * t ** 2 + a4 * t ** 3 + a5 / t + a6 * ln(t)
    return parameter


def parameter_cal_moller(a1, a2, a3, a4, a5, a6, a7, a8, t):
    p = a1 + a2 * t + a3 / t + a4 * ln(t) + a5 / (t - 263) + a6 * t ** 2 + a7 / (680 - t) + a8 / (t - 227)
    return p

def parameter_cal_holmes(a1, a2, a3, a4, a5, a6, t):
    t_r = 298.15
    parameter = a1 + a2 * (1 / t - 1 / t_r) + a3 * ln(t / t_r) + a4 * (t - t_r) + a5 * (t ** 2 - t_r ** 2) + a6 * ln(
        t - 260)
    return parameter

def parameter_cal_marion(a1, a2, a3, a4, a5, a6, a7, t):
    # reference: [4]
    parameter = a1 + a2 * t + a3 * t ** 2 + a4 * t ** 3 + a5 / t + a6 * ln(t) + a7 / (t ** 2)
    return parameter


def get_parameter_lassin(name, data, t):
    """
    Choose equation and calculate parameters conveniently.
    :param name: parameter name, can be 'b0', 'b1', 'c', 'theta', 'psi'.
    :param data: a dict contains all the interaction data an ion-pair.
    :param t: temperature of the solution.
    :param method: method used to calculate parameter, default is 'spencer'.
    :return: value of the parameter.
    """
    # print(data)
    a1 = data['{}_a1'.format(name)]
    a2 = data['{}_a2'.format(name)]
    a3 = data['{}_a3'.format(name)]
    a4 = data['{}_a4'.format(name)]
    a5 = data['{}_a5'.format(name)]
    a6 = data['{}_a6'.format(name)]
    a7 = data['{}_a7'.format(name)]
    a8 = data['{}_a8'.format(name)]

    parameter = parameter_cal_lassin(a1, a2, a3, a4, a5, a6, a7, a8, t)

    return parameter


def get_parameter_spencer(name, data, t):
    """
    Choose equation and calculate parameters conveniently.
    :param name: parameter name, can be 'b0', 'b1', 'c', 'theta', 'psi'.
    :param data: a dict contains all the interaction data an ion-pair.
    :param t: temperature of the solution.
    :param method: method used to calculate parameter, default is 'spencer'.
    :return: value of the parameter.
    """

    a1 = data['{}_a1'.format(name)]
    a2 = data['{}_a2'.format(name)]
    a3 = data['{}_a3'.format(name)]
    a4 = data['{}_a4'.format(name)]
    a5 = data['{}_a5'.format(name)]
    a6 = data['{}_a6'.format(name)]

    parameter = parameter_cal_spencer(a1, a2, a3, a4, a5, a6, t)

    return parameter

def get_parameter_marion(name, data, t):
    a1 = data['{}_a1'.format(name)]
    a2 = data['{}_a2'.format(name)]
    a3 = data['{}_a3'.format(name)]
    a4 = data['{}_a4'.format(name)]
    a5 = data['{}_a5'.format(name)]
    a6 = data['{}_a6'.format(name)]
    a7 = data['{}_a7'.format(name)]

    parameter = parameter_cal_marion(a1, a2, a3, a4, a5, a6, a7, t)
    return parameter

def get_parameter_holmes(name, data, t):
    """
    Choose equation and calculate parameters conveniently.
    :param name: parameter name, can be 'b0', 'b1', 'c', 'theta', 'psi'.
    :param data: a dict contains all the interaction data an ion-pair.
    :param t: temperature of the solution.
    :param method: method used to calculate parameter, default is 'spencer'.
    :return: value of the parameter.
    """
    # print(data)
    a1 = data['{}_a1'.format(name)]
    a2 = data['{}_a2'.format(name)]
    a3 = data['{}_a3'.format(name)]
    a4 = data['{}_a4'.format(name)]
    a5 = data['{}_a5'.format(name)]
    a6 = data['{}_a6'.format(name)]


    parameter = parameter_cal_holmes(a1, a2, a3, a4, a5, a6, t)

    return parameter


def get_parameter_moller(name, data, t):
    """
    Choose equation and calculate parameters conveniently.
    :param name: parameter name, can be 'b0', 'b1', 'c', 'theta', 'psi'.
    :param data: a dict contains all the interaction data an ion-pair.
    :param t: temperature of the solution.
    :param method: method used to calculate parameter, default is 'spencer'.
    :return: value of the parameter.
    """
    # print(data)
    a1 = data['{}_a1'.format(name)]
    a2 = data['{}_a2'.format(name)]
    a3 = data['{}_a3'.format(name)]
    a4 = data['{}_a4'.format(name)]
    a5 = data['{}_a5'.format(name)]
    a6 = data['{}_a6'.format(name)]
    a7 = data['{}_a7'.format(name)]
    a8 = data['{}_a8'.format(name)]

    parameter = parameter_cal_moller(a1, a2, a3, a4, a5, a6, a7, a8, t)
    return parameter


def get_parameter(pair, name, data, t):
    if 'Li+' in pair or 'Cs+' in pair:
        return get_parameter_holmes(name, data, t)
    elif 'Fe+2' in pair:
        return get_parameter_marion(name,data,t)
    else:
        return get_parameter_spencer(name, data, t)


def get_charge_number(ion):
    """
    get the charge number of an ion (str)
    :param ion: ion name, a string with "+" or "-" sign followed by number of charge
    :return: charge number
    """
    if "+" in ion:
        lis = ion.split("+")
        if lis[1]:
            result = lis[1]
        else:
            result = 1
    elif "-" in ion:
        lis = ion.split("-")
        if lis[1]:
            lis[1] = '-' + lis[1]
            result = lis[1]
        else:
            result = -1
    else:
        result = 0
    return int(result)


def is_neutral(ion):
    charge = get_charge_number(ion)
    if charge == 0:
        return True
    return False


def is_cation(ion):
    charge = get_charge_number(ion)
    if charge > 0:
        return True
    return False


def is_anion(ion):
    charge = get_charge_number(ion)
    if charge < 0:
        return True
    return False


def species_type(species):
    """
    Determine whether a species is a cation, anion or neutral species.
    :param species: [string], e.g. "Na+", "Cl-", "H2O".
    :return: type name
    """
    if species.find("+") != -1:
        s_type = 'cation'
    elif species.find("-") != -1:
        s_type = 'anion'
    else:
        s_type = 'neutral'
    return s_type


def salt_type(pair):
    """
    Determine whether a salt is a 2-2 type or not.
    :param ion1: ion1
    :param ion2: ion2
    :return: return True if it is a 2-2 type of salt
    """
    ion1 = pair[0]
    ion2 = pair[1]
    if ion1.find("+") != -1:
        result1 = ion1.split("+")
    else:
        result1 = ion1.split("-")
    if ion2.find("+") != -1:
        result2 = ion2.split("+")
    else:
        result2 = ion2.split("-")
    if result1[1] == '2' and result2[1] == '2':
        return True
    else:
        return False


"""
the Pitzer parameters
"""


def a_phi_moller(t):
    a1 = 3.36901532e-01
    a2 = -6.32100430e-04
    a3 = 9.14252359e00
    a4 = -1.35143986e-02
    a5 = 2.26089488e-03
    a6 = 1.92118597e-06
    a7 = 4.52586464e01
    a8 = 0
    return a1 + a2 * t + a3 / t + a4 * ln(t) + a5 / (t - 263) + a6 * t ** 2 + a7 / (680 - t) + a8 / (t - 227)


def a_phi_spencer(t):
    """
    Calculate the A_phi (Debye-Hukel constant) according to the temperature
    :param t: temperature of solution, Calvin(K)
    :return: value of a_phi
    :reference:
    """
    a1 = 8.66836498e1
    a2 = 8.48795942e-2
    a3 = -8.88785150e-5
    a4 = 4.88096393e-8
    a5 = -1.32731477e3
    a6 = -1.76460172e1
    a_phi = parameter_cal_spencer(a1, a2, a3, a4, a5, a6, t)
    return a_phi


def g_func(a):
    result = 2 * (
            1 - (1 + a) * exp(-a)
    ) / a ** 2
    return result


def g_func_prime(a):
    g_prime = -2 * (
            1 - (1 + a + a ** 2 / 2) * exp(-a)
    ) / a ** 2
    return g_prime


def get_beta(parameters, pair, i):
    charge_number1 = get_charge_number(pair[0])
    charge_number2 = get_charge_number(pair[1])

    alpha = 2  # kg^(1/2)⋅mol^(-1/2)
    alpha_1 = 1.4  # kg^(1/2)⋅mol^(-1/2)
    alpha_2 = 12  # kg^(1/2)⋅mol^(-1/2)
    b0 = parameters['b0']
    b1 = parameters['b1']
    b2 = parameters['b2']

    if abs(charge_number1) == 2 and abs(charge_number2) == 2:
        beta_mx = b0 + b1 * g_func(
            alpha_1 * (i ** (1 / 2))
        ) + b2 * g_func(
            alpha_2 * (i ** (1 / 2))
        )
    else:
        beta_mx = b0 + b1 * g_func(
            alpha * i ** (1 / 2)
        )
    return beta_mx


def get_beta_prime(parameters, pair, i):
    charge_number1 = get_charge_number(pair[0])
    charge_number2 = get_charge_number(pair[1])

    alpha = 2  # kg^(1/2)⋅mol^(-1/2)
    alpha_1 = 1.4  # kg^(1/2)⋅mol^(-1/2)
    alpha_2 = 12  # kg^(1/2)⋅mol^(-1/2)
    b0 = parameters['b0']
    b1 = parameters['b1']
    b2 = parameters['b2']

    if abs(charge_number1) == 2 and abs(charge_number2) == 2:
        beta_prime = (
                             b1 * g_func_prime(
                         alpha_1 * (i ** (1 / 2))
                     ) + b2 * g_func_prime(
                         alpha_2 * (i ** (1 / 2))
                     )
                     ) / i
    else:
        beta_prime = b1 * g_func_prime(alpha * i ** (1 / 2)) / i
    return beta_prime


def get_beta_phi(parameters, pair, i):
    charge_number1 = get_charge_number(pair[0])
    charge_number2 = get_charge_number(pair[1])

    alpha = 2  # kg^(1/2)⋅mol^(-1/2)
    alpha_1 = 1.4  # kg^(1/2)⋅mol^(-1/2)
    alpha_2 = 12  # kg^(1/2)⋅mol^(-1/2)

    b0 = parameters['b0']
    b1 = parameters['b1']
    b2 = parameters['b0']

    """for 2-2 type salts"""
    if abs(charge_number1) == 2 and abs(charge_number2) == 2:
        beta_phi = b0 + b1 * exp(-alpha_1 * (i ** (1 / 2))) + b2 * exp(-alpha_2 * (i ** (1 / 2)))
    else:
        """for 1-1, 1-2, 2-1, 3-1, 4-1 type salts"""
        beta_phi = b0 + b1 * exp(-alpha * (i ** (1 / 2)))
    return beta_phi


def get_c(c_phi, z_m, z_x):
    c_mx = c_phi / (2 * (abs(z_m * z_x)) ** 0.5)
    return c_mx


def get_c_gamma(c_phi):
    c_gamma = 3 * c_phi / 2
    return c_gamma


def get_f(a_phi, i):
    """
    :param a_phi: A_phi (Debye-Hukel constant)
    :param i: ionic strength
    :return: expression of "f" function
    """
    b = 1.2
    f = - (4 * i * a_phi / b) * ln(1 + b * i ** (1 / 2))
    return f


def get_f_gamma(a_phi, i):
    """
    :param a_phi:A_phi (Debye-Hukel constant)
    :param i: ionic strength
    :return: the "f^gamma" function in Pitzer's model
    """
    b = 1.2
    f_gamma = - a_phi * (
            i ** (1 / 2) / (1 + b * i ** (1 / 2)) + (2 / b) * ln(
        1 + b * i ** (1 / 2))
    )
    return f_gamma


"""
*** End of the Pitzer parameters dealing *** 
"""


def get_x_mn(z_m: int, z_n: int, a_phi: float, i: float):
    """
    calculate the 'x' value of ions 'm' and 'n'.
    :param z_m: charge number of ion 'm'
    :param z_n: charge number of ion 'n'
    :param a_phi: Avogedral's number of this solution
    :param i: Ionic strength of this solution
    :return: 'x_mn' for calculating the 'J' value
    :reference: [2] p9
    """
    x_mn = 6 * z_m * z_n * a_phi * i ** 0.5
    return x_mn


def get_e_theta(z_m, z_n, a_phi, i):
    """
    :param z_m: charge number of species m
    :param z_n: charge number of species n
    :param a_phi:
    :param i: ionic strength
    :return: e_theta and e_theta_prime
    :reference: [1] p123
    """

    x_mn = get_x_mn(z_m, z_n, a_phi, i)
    x_mm = get_x_mn(z_m, z_m, a_phi, i)
    x_nn = get_x_mn(z_n, z_n, a_phi, i)

    mn = compute_j_jp(x_mn)
    mm = compute_j_jp(x_mm)
    nn = compute_j_jp(x_nn)
    j_mn = mn['j_x']
    j_mn_prime = mn['j_xp']
    j_mm = mm['j_x']
    j_mm_prime = mm['j_xp']
    j_nn = nn['j_x']
    j_nn_prime = nn['j_xp']

    e_theta = (z_m * z_n / (4 * i)) * (j_mn - 0.5 * j_mm - 0.5 * j_nn)
    e_theta_prime = -(e_theta / i) + (z_m * z_n / (8 * i ** 2)) * (
            x_mn * j_mn_prime - 0.5 * x_mm * j_mm_prime - 0.5 * x_nn * j_nn_prime)
    return {
        "e_theta": e_theta,
        "e_theta_prime": e_theta_prime,
    }


@hash_dict
@lru_cache(maxsize=None)
def calculate_ionic_strength(molalities):
    data = molalities
    ions = data.keys()
    sum_value = 0
    for ion in ions:
        charge_number = get_charge_number(ion)
        sum_value += data[ion] * (charge_number ** 2)
    return sum_value / 2


@hash_dict
@lru_cache(maxsize=None)
def calculate_molality(x, species):
    x1, x2 = x
    molalities = {}
    for key, value in species.items():
        molalities[key] = value * x1

    if 'Cl-' not in species.keys():
        molalities['Cl-'] = x2
    return molalities


@hash_dict
@lru_cache(maxsize=None)
def calculate_charge_balance(x, molalities):
    balance = 0
    for species in molalities.keys():
        balance += get_charge_number(species) * molalities[species]
    return balance


@lru_cache(maxsize=None)
def get_chemical_potential(species, t):
    """
    Calculate the standard chemical potential of solids melting reaction.
    the "chemical potential" here actually means "μ/RT".
    :param solid: [string], solid species, e.g. "NaCl"
    :param t: [number], melting temperature of the solid, in Kelvin.
    :return: [number], standard chemical potential of the melting reaction.
    """
    if 'Fe+2' in species:
        data = marion_chemical_potential_db.loc[species]
        std_chemical_potential = parameter_cal_marion(
            data['a1'],
            data['a2'],
            data['a3'],
            data['a4'],
            data['a5'],
            data['a6'],
            data['a7'],
            t
        )
    elif species == 'LiCl0':
        data = lassin_chemical_potential_db.loc[species]
        std_chemical_potential = chemical_potential_lassin(
            data['a1'],
            data['a2'],
            data['a3'],
            data['a4'],
            data['a5'],
            t
        )
    else:
        data = spencer_chemical_potential_db.loc[species]
        std_chemical_potential = parameter_cal_spencer(
            a1=data['a1'],
            a2=data['a2'],
            a3=data['a3'],
            a4=data['a4'],
            a5=data['a5'],
            a6=data['a6'],
            t=t
        )
    return std_chemical_potential


def get_hydrate_data(solid):
    data = {}
    if solid in solids.keys():
        data = solids[solid]
    return data


@hash_dict
@lru_cache(maxsize=None)
def group_components(components):
    """
    Find groups from components of ions and neutral species
    :param components: consists of cations, anions and neutral species.
    :return: groups
    """
    cations = [c for c in components if '+' in c]
    anions = [a for a in components if '-' in a]
    neutrals = [n for n in components if '+' not in n and '-' not in n]

    cation_anion_pairs = list(itertools.product(cations, anions))
    cation_pairs = list(itertools.combinations(cations, 2)) if len(cations) >= 2 else []
    anion_pairs = list(itertools.combinations(anions, 2)) if len(anions) >= 2 else []
    neutral_pairs = list(itertools.combinations(neutrals, 2))

    neutral_ion_pairs = list(itertools.product(neutrals, cations + anions))

    neutral_cation_anion_pairs = [(a, *b) for a in neutrals for b in cation_anion_pairs]

    return {
        'cations': cations,
        'anions': anions,
        'neutrals': neutrals,
        'cation_anion_pairs': cation_anion_pairs,
        'cation_pairs': cation_pairs,
        'anion_pairs': anion_pairs,
        'neutral_pairs': neutral_pairs,
        'neutral_ion_pairs': neutral_ion_pairs,
        'neutral_cation_anion_pairs': neutral_cation_anion_pairs
    }


"""
Pitzer-model-only methods
"""


@hash_dict
@lru_cache(maxsize=None)
def get_beta_012(rd, t, method):
    b0 = 0
    b1 = 0
    b2 = 0
    if method == 'lassin':
        b0 = get_parameter_lassin(name='b0', data=rd, t=t)
        b1 = get_parameter_lassin(name='b1', data=rd, t=t)
        b2 = get_parameter_lassin(name='b2', data=rd, t=t)
    if method == 'spencer':
        b0 = get_parameter_spencer(name='b0', data=rd, t=t)
        b1 = get_parameter_spencer(name='b1', data=rd, t=t)
        b2 = get_parameter_spencer(name='b2', data=rd, t=t)
    if method == 'marion':
        b0 = get_parameter_marion(name='b0', data=rd, t=t)
        b1 = get_parameter_marion(name='b1', data=rd, t=t)
        b2 = get_parameter_marion(name='b2', data=rd, t=t)
    elif method == 'moller':
        b0 = get_parameter_moller(name='b0', data=rd, t=t)
        b1 = get_parameter_moller(name='b1', data=rd, t=t)
        b2 = get_parameter_moller(name='b2', data=rd, t=t)
    elif method == 'holmes':
        b0 = get_parameter_holmes(name='b0', data=rd, t=t)
        b1 = get_parameter_holmes(name='b1', data=rd, t=t)
        b2 = get_parameter_holmes(name='b2', data=rd, t=t)
    return {
        'b0': b0,
        'b1': b1,
        'b2': b2,
    }


@hash_dict
@lru_cache(maxsize=None)
def beta_calculate(ion_pair, ionic_strength, t, database):
    rd = binary_parameters_ready(ion_pair, t, database)
    if 'Li+' in ion_pair or 'Cs+' in ion_pair:
        method = 'holmes'
    elif 'Fe+2' in ion_pair:
        method = 'marion'
    else:
        method = 'spencer'
    beta_012 = get_beta_012(rd, t, method=method)

    beta = get_beta(beta_012, ion_pair, ionic_strength)
    return beta


@hash_dict
@lru_cache(maxsize=None)
def beta_phi_calculate(ion_pair, ionic_strength, t, database):
    rd = binary_parameters_ready(ion_pair, t, database)
    if 'Li+' in ion_pair or 'Cs+' in ion_pair:
        method = 'holmes'
    else:
        method = 'spencer'
    beta_012 = get_beta_012(rd, t, method=method)
    b_phi = get_beta_phi(beta_012, ion_pair, ionic_strength)

    return b_phi


@hash_dict
@lru_cache(maxsize=None)
def beta_prime_calculate(ion_pair, ionic_strength, t, database):
    rd = binary_parameters_ready(ion_pair, t, database)

    if 'Li+' in ion_pair or 'Cs+' in ion_pair:
        method = 'holmes'
    else:
        method = 'spencer'
    beta_012 = get_beta_012(rd, t, method=method)
    b_prime = get_beta_prime(beta_012, ion_pair, ionic_strength)
    return b_prime


@hash_dict
@lru_cache(maxsize=None)
def c_calculate(ion_pair, t, database):
    rd = binary_parameters_ready(ion_pair, t, database)
    if 'Li+' in ion_pair or 'Cs+' in ion_pair:
        c0 = get_parameter_holmes(name='c_phi', data=rd, t=t)
    else:
        c0 = get_parameter_spencer(name='c_phi', data=rd, t=t)
    charge_number1 = get_charge_number(ion_pair[0])
    charge_number2 = get_charge_number(ion_pair[1])
    c = get_c(c0, charge_number1, charge_number2)
    return c


@hash_dict
@lru_cache(maxsize=None)
def cc_phi_calculate(ion_pair, a_phi, ionic_strength, t, database):
    charge_number1 = get_charge_number(ion_pair[0])
    charge_number2 = get_charge_number(ion_pair[1])
    rd = binary_parameters_ready(ion_pair, t, database)

    theta = get_parameter_spencer(name='theta', data=rd, t=t)

    if charge_number1 != charge_number2:
        e_thetas = get_e_theta(charge_number1, charge_number2, a_phi, ionic_strength)
        e_theta = e_thetas['e_theta']
        e_theta_prime = e_thetas['e_theta_prime']
    else:
        e_theta = 0
        e_theta_prime = 0
    phi = theta + e_theta
    phi_prime = e_theta_prime

    return {
        'phi': phi,
        'phi_prime': phi_prime
    }


@hash_dict
@lru_cache(maxsize=None)
def aa_phi_calculate(ion_pair, a_phi, ionic_strength, t, database):
    charge_number1 = get_charge_number(ion_pair[0])
    charge_number2 = get_charge_number(ion_pair[1])

    rd = binary_parameters_ready(ion_pair, t, database)

    if 'Li+' in ion_pair or 'LiCl0' in ion_pair:
        theta = get_parameter_lassin(name='theta', data=rd, t=t)
    else:
        theta = get_parameter_spencer(name='theta', data=rd, t=t)
    if charge_number1 != charge_number2:
        e_thetas = get_e_theta(charge_number1, charge_number2, a_phi, ionic_strength)
        e_theta = e_thetas['e_theta']
        e_theta_prime = e_thetas['e_theta_prime']
    else:
        e_theta = 0
        e_theta_prime = 0
    phi = theta + e_theta
    phi_prime = e_theta_prime

    return {
        "phi": phi,
        "phi_prime": phi_prime
    }


"""
Reference
[1] Pitzer K S. Activity coefficients in electrolyte solutions[M]. CRC press, 2018.
[2] Pitzer K S. Thermodynamics of electrolytes. V. Effects of higher-order electrostatic terms[J]. Journal of Solution 
Chemistry, 1975, 4(3): 249-265.
[3] Bradley, Daniel J., and Kenneth S. Pitzer. "Thermodynamics of electrolytes. 12. Dielectric properties of water and 
Debye-Hueckel parameters to 350. degree. C and 1 kbar." Journal of physical chemistry 83.12 (1979): 1599-1603.
[4] Marion GM, Catling DC, Kargel JS. Modeling aqueous ferrous iron chemistry at low temperatures with application to 
Mars. Geochimica et cosmochimica Acta. 2003 Nov 15;67(22):4251-66.
"""



# -*- coding: utf-8 -*-
# Author: Yiping Liu
# Description: This script calculates the average of a list of numbers.
# Version: 1.0
# Last Modified: May 7, 2023

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

# read the csv file
df = pd.read_csv("E:\\work\\data\\low-T\\error_analysis.csv")

# extract the columns of interest
RK = df['RK']
Tm = df['Tm']
mNa = df['mNa']

# plot the histograms
fig, axs = plt.subplots(1, 3, figsize=(12, 4))
axs[0].hist(RK, alpha=0.5, edgecolor='black', bins=np.linspace(RK.min(), RK.max(), 30))
axs[0].set_title("RK")
axs[1].hist(Tm, alpha=0.5, edgecolor='black', bins=np.linspace(Tm.min(), Tm.max(), 30))
axs[1].set_title("Tm")
axs[2].hist(mNa, alpha=0.5, edgecolor='black', bins=np.linspace(mNa.min(), mNa.max(), 30))
axs[2].set_title("mNa")

# find the 95% confidence interval for the last histogram
n, bins, patches = axs[2].hist(mNa, alpha=0.5, edgecolor='black', bins=np.linspace(mNa.min(), mNa.max(), 30))
bin_centers = 0.5 * (bins[:-1] + bins[1:])
cdf = np.cumsum(n)
cdf = cdf / cdf[-1]
lower_bound = np.interp(0.05, cdf, bin_centers)
upper_bound = np.interp(0.95, cdf, bin_centers)
conf_interval = (lower_bound, upper_bound)
print("95% confidence interval for mNa histogram:", conf_interval)

# add a transparent gray rectangle to represent the 95% confidence interval in the last histogram
for patch, left, right in zip(patches, bins[:-1], bins[1:]):
    if left <= lower_bound or right >= upper_bound:
        patch.set_alpha(0.5)
        patch.set_fc('gray')

# save the plot to pdf
plt.tight_layout()
plt.savefig("E:\\work\\plots\\error_histogram.pdf")
plt.show()